/*
 * mpgedit playback plugin using MPGLIB decoder library
 *
 * Copyright (C) 2001 Adam Bernstein
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307  USA.
 */

#ifndef lint
static char SccsId[] = "$Id: decoder_mpg123.c,v 1.7 2001/10/03 14:20:43 number6 Exp $";
#endif

/*
 * API to access native sound card interface
 */
#include "dsp.h"

/*
 * mpglib interface header
 */
#include <interface.h>
#include <stdlib.h>
#include "decoder.h"

#define MPGDECODER_BUFSIZ (5 * 1024)
#define MPGDECODER_DEFAULT_SAMPLERATE 44100
#define MPGDECODER_SOUNDDEV "/dev/dsp"

plotting_data *mpg123_pinfo = NULL;

static void _mpgdecoder_free(mpgdecoder_ctx *ctx)
{
    mpgdecoder_ctx *mpctx = (mpgdecoder_ctx *) ctx;
    if (mpctx) {
        if (mpctx->g_outmpbuf) {
            free(mpctx->g_outmpbuf);
        }
        if (mpctx->g_mpctx) {
            free(mpctx->g_mpctx);
        }
        free(mpctx);
    }
}


void *mpgdecoder_alloc(void) 
{
    mpgdecoder_ctx *newctx = NULL;
    int sts = 1;
    
    newctx = calloc(1, sizeof(mpgdecoder_ctx));
    if (!newctx) {
        goto clean_exit;
    }

    /*
     * Hack:  This code does not work right when we allocate
     * just sizeof(MPSTR). Something in mpglib is using more
     * memory than it is supposed to.  The hack here is to 
     * allocate 2x what is needed.
     */
    newctx->g_mpctx = calloc(2, sizeof(MPSTR));
    if (!newctx->g_mpctx) {
        goto clean_exit;
    }
    newctx->g_outmpbuf = calloc(1, MPGDECODER_BUFSIZ);
    if (!newctx->g_outmpbuf) {
        goto clean_exit;
    }
    newctx->g_outmpbuflen = MPGDECODER_BUFSIZ;
    sts = 0;

clean_exit:
    if (sts) {
        /*
         * Alloc error occurred; free everything
         */
        _mpgdecoder_free(newctx);
        newctx = NULL;
    }
    return newctx;
}


void mpgdecoder_free(void *ctx)
{
    _mpgdecoder_free((mpgdecoder_ctx *) ctx);
}


void mpgdecoder_init(void *ctx)
{
    mpgdecoder_ctx *mpctx = (mpgdecoder_ctx *) ctx;
    if (mpctx && mpctx->g_mpctx) {
        InitMP3((MPSTR *) mpctx->g_mpctx);
    }
}


FILE *mpgdecoder_open(void *ctx, int sample_rate, int stereo)
{
    mpgdecoder_ctx *mpctx = (mpgdecoder_ctx *) ctx;
    FILE           *playfp;

    freopen("/dev/null", "a+", stderr);
    playfp = fopen(MPGDECODER_SOUNDDEV, "w");
    if (playfp) {
        /*
         * Files without a valid mp3 header at the start of 
         * the file report a 0 sample rate.  Assume 44100 
         * for such files.
         */
        if (sample_rate == 0) {
            sample_rate = MPGDECODER_DEFAULT_SAMPLERATE;
        }
        init_dsp(fileno(playfp), sample_rate, stereo);
        mpctx->g_sample_rate = sample_rate;
    }

    return playfp;
}


void mpgdecoder_play_frame(void          *ctx,
                           FILE          **playfp,
                           unsigned char *buf,
                           int           len,
                           int           rate,
                           int           stereo)
{
    mpgdecoder_ctx *mpctx = (mpgdecoder_ctx *) ctx;
    int            done = 0;
    int            iret = 0;

    iret = decodeMP3(mpctx->g_mpctx, buf, len, 
                     mpctx->g_outmpbuf, mpctx->g_outmpbuflen, &done);
    if (iret == MP3_OK) {
        if (rate != mpctx->g_sample_rate) {
            /*
             * Really want to just change the sound card playback sample rate.
             * However, must flush all pending data first, by closing
             * playback device.  It would not be good to change the
             * sample rate while data with the previous sample rate is
             * still in the sound card buffers.
             */
            fclose(*playfp);
            InitMP3((MPSTR *) mpctx->g_mpctx);
            *playfp = mpgdecoder_open(ctx, rate, stereo);
        }
        fwrite(mpctx->g_outmpbuf, done, 1, *playfp);
    }
}
