/*
 * soundcard initialization function for decoder_mpg123 playback plugin
 *
 * Copyright (C) 2001 Adam Bernstein
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307  USA.
 */

#ifndef lint
static char SccsId[] = "$Id: dsp.c,v 1.3 2001/10/03 14:20:44 number6 Exp $";
#endif

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/soundcard.h>
#include <stdio.h>

int init_dsp(int fd, int sample_rate, int stereo)
{
    int retval = 0;
    int format = 0;
    int val;

    format = AFMT_S16_LE;
    val = format;
    if (ioctl(fd, SNDCTL_DSP_SETFMT, &val) == -1)  {
        perror("SNDCTL_DSP_SETFMT");
        retval = 1;
        goto clean_exit;

    }

    /*
     * Stereo format
     */
    val    = stereo;     /* 0=mono, 1=stereo */
    format = val;
    if (ioctl(fd, SNDCTL_DSP_STEREO, &format) == -1) {
        perror("SNDCTL_DSP_STEREO");
        retval = 1;
        goto clean_exit;
    }

    if (format != val) {
        printf("failed to set stereo format...  %d %d bailing\n", val, format);
        retval = 1;
        goto clean_exit;
    }

    /*
     * Sample rate
     */
    val = sample_rate;
    format = val;
    if (ioctl(fd, SNDCTL_DSP_SPEED, &format) == -1)  {
        perror("SNDCTL_DSP_SPEED");
        retval = 1;
        goto clean_exit;
    }
    if (format != val) {
        /*
         * Not fatal if a non-standard sample was specified, but
         * do print a warning message.
         */
        printf("Warning: failed setting %d sample size... %d\n", val, format);
        goto clean_exit;
    }


clean_exit:
    return retval;
}
