/*
 * MPEG audio file time indexing routines
 *
 * Copyright (C) 2001 Adam Bernstein
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307  USA.
 */

#ifndef lint
static char SccsId[] = "$Id: mpegindx.c,v 1.5 2001/10/22 22:49:51 number6 Exp $";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "mpegfio.h"
#include "header.h"
#include "mp3time.h"
#include "mp3_header.h"
#include "xing_header.h"

int mp3edit_skip_xing_header(FILE *fp)
{
    unsigned char    xing_buf[2048];
    int              cnt;
    int              has_xing = 0;
    XHEADDATA        xingh;
    mpeg_header_data mpeg_header;
    unsigned char    mpeg_header_data[4];

    memset(&xingh,   0, sizeof(xingh));
    memset(xing_buf, 0, sizeof(xing_buf));
    memset(&mpeg_header, 0, sizeof(mpeg_header));

    cnt = fread(mpeg_header_data, sizeof(mpeg_header_data), 1, fp);
    if (cnt != 1) {
        return 0;
    }

    cnt = decode_mpeg_header(mpeg_header_data, &mpeg_header, 0);
    if (!cnt) {
        return 0;
    }

    memcpy(xing_buf, mpeg_header_data, sizeof(mpeg_header_data));
    cnt = fread(&xing_buf[4], mpeg_header.frame_size - 4, 1, fp);
    if (cnt == 1) {
        has_xing =  get_xing_header(xing_buf, &xingh);
    }
    else {
        return 0;
    }

    if (!has_xing) {
        rewind(fp);
    }
    return has_xing ? mpeg_header.frame_size : 0;
}


void mp3edit_create_indexfile(FILE *fp, FILE *indxfp,
                              mpegfio_iocallbacks *ttyio)
{
    mpeg_header_data header;
    mpeg_time        mpegtval;
    mpeg_file_iobuf  mpegiobuf;
    int              tlast = 0;
    int              tnow  = 0;
    long             file_position = 0;
    int              found = 0;
    int              frame_cnt = 0;
    int              eof = 0;

    memset(&header, 0, sizeof(header));
    mpeg_time_clear(&mpegtval);
    mpeg_file_iobuf_clear(&mpegiobuf);

    /*
     * Second 0 corresponds to the start of the file, not counting the
     * Xing header.
     */
    fwrite(&file_position, sizeof(file_position), 1, indxfp);

    /*
     * Skip over Xing header for time counting purposes.
     */
    file_position = mp3edit_skip_xing_header(fp);

    do {
        found = read_next_mpeg_frame(fp, &mpegiobuf, &header, &eof);
        if (found == 1) {
            frame_cnt++;
            mpeg_time_frame_increment(&mpegtval, &header);
            tnow = mpeg_time_2seconds(&mpegtval);
            file_position += header.position;
            file_position += header.frame_size;

            
            if (tnow > tlast) {
                tlast = tnow;
                fwrite(&file_position, sizeof(file_position), 1, indxfp);
                if (ttyio && ttyio->printf) {
                    ttyio->printf(ttyio->ctx,
                        "\rElapsed time: %4d:%02d| %5ds|Frame: %-7d\r",
                           tnow/60,
                           tnow%60,
                           tnow,
                           frame_cnt);
                }
            }
            mpeg_file_iobuf_setstart(&mpegiobuf,
                                    mpeg_file_iobuf_getstart(&mpegiobuf) +
                                         header.frame_size);
        }
        /* found */
    } while (!eof);
}
